#
# debug.py
#
# Copyright (c) 2017-2025 Junpei Kawamoto
#
# This file is part of rgmining-tripadvisor-dataset.
#
# rgmining-tripadvisor-dataset is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# rgmining-tripadvisor-dataset is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Foobar.  If not, see <http://www.gnu.org/licenses/>.
#
"""This module provides a debug function for the Trip Advisor Dataset."""

import json
import sys
from typing import Protocol, TextIO, Any, TypeVar


class Reviewer(Protocol):
    """A protocol for a reviewer object."""

    @property
    def name(self) -> str:
        """The reviewer's ID.'"""

    @property
    def anomalous_score(self) -> float:
        """The anomalous score of the reviewer."""


class Product(Protocol):
    """A protocol for a product object."""

    @property
    def name(self) -> str:
        """The product's ID.'"""

    @property
    def summary(self) -> Any:
        """The summary of the reviews for the product."""


RT = TypeVar("RT", bound=Reviewer)
PT = TypeVar("PT", bound=Product)


class Graph(Protocol[RT, PT]):
    """A protocol for a graph object."""

    @property
    def reviewers(self) -> list[RT]:
        """A list of reviewers."""

    @property
    def products(self) -> list[PT]:
        """A list of products."""


def print_state(g: Graph, i: int | str, output: TextIO = sys.stdout) -> None:
    """Print a current state of a given graph.

    This method outputs a current of a graph as a set of json objects.
    Graph objects must have two properties, `reviewers` and `products`.
    Those properties return a set of reviewers and products respectively.
    See the :ref:`graph interface <dataset-io:graph-interface>`
    for more information.

    In this output format, each line represents a reviewer or product object.

    Reviewer objects are defined as ::

        {
           "iteration": <the iteration number given as i>
           "reviewer":
           {
              "reviewer_id": <Reviewer's ID>
              "score": <Anomalous score of the reviewer>
           }
        }

    Product objects are defined as ::

        {
           "iteration": <the iteration number given as i>
           "reviewer":
           {
              "product_id": <Product's ID>
              "summary": <Summary of the reviews for the product>
           }
        }

    Args:
      g: Graph instance.
      i: Iteration number.
      output: A writable object (default: sys.stdout).
    """
    for r in g.reviewers:
        json.dump(
            {
                "iteration": i,
                "reviewer": {
                    "reviewer_id": r.name,
                    "score": r.anomalous_score,
                },
            },
            output,
        )
        output.write("\n")

    for p in g.products:
        json.dump(
            {
                "iteration": i,
                "product": {
                    "product_id": p.name,
                    "summary": float(str(p.summary)),
                },
            },
            output,
        )
        output.write("\n")



#
# cli.py
#
# Copyright (c) 2017-2025 Junpei Kawamoto
#
# This file is part of rgmining-tripadvisor-dataset.
#
# rgmining-tripadvisor-dataset is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# rgmining-tripadvisor-dataset is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Foobar.  If not, see <http://www.gnu.org/licenses/>.
#
"""Evaluate a review graph mining algorithm with the Trip Advisor dataset.

Usage: python -m tripadvisor [OPTIONS]

  Evaluate a review graph mining algorithm with the Trip Advisor dataset.

Options:
  -m, --method [rsd|feagle|fraudar]
                                  name of algorithm.  [required]
  --loop INTEGER                  number of iteration.
  --threshold FLOAT               threshold.
  --output FILENAME               file path to store results. [Default:
                                  stdout]
  --param TEXT                    key and value a pair of parameters
                                  corresponding to the chosen algorithm,
                                  connected with '='.
  --version                       Show the version and exit.
  --help                          Show this message and exit.
"""

import logging
import sys
from importlib.metadata import version
from typing import TextIO, Callable, Any, Protocol, Final

import click

from tripadvisor.debug import print_state, Graph as PrintableGraph
from tripadvisor.loader import load, Graph as LoadableGraph

LOGGER = logging.getLogger(__name__)


class Graph(PrintableGraph, LoadableGraph, Protocol):
    def update(self) -> float: ...


ALGORITHMS = dict[str, Callable[..., Graph]]()
"""Dictionary of graph loading functions associated with installed algorithms.
"""

# Load and register RIA.
try:
    import ria

    def ignore_args(func: Callable) -> Callable:
        """Returns a wrapped function which ignore given arguments."""

        def _(*_args: tuple[Any]) -> Any:
            """The function body."""
            return func()

        return _

    ALGORITHMS["ria"] = ria.ria_graph
    ALGORITHMS["one"] = ignore_args(ria.one_graph)
    ALGORITHMS["onesum"] = ignore_args(ria.one_sum_graph)
    ALGORITHMS["mra"] = ignore_args(ria.mra_graph)
except ImportError:
    LOGGER.info("rgmining-ria is not installed.")

# Load and register RSD.
try:
    import rsd

    def create_rsd_graph(**kwargs: float) -> rsd.ReviewGraph:
        """Create a review graph defined in RSD package."""
        if "theta" not in kwargs:
            LOGGER.warning("Parameter 'theta' is not specified. Set to 0.1.")
            kwargs["theta"] = 0.1
        return rsd.ReviewGraph(kwargs["theta"])

    ALGORITHMS["rsd"] = create_rsd_graph
except ImportError:
    LOGGER.info("rgmining-rsd is not installed.")

# Load and register Fraud Eagle.
try:
    import fraud_eagle

    def create_feagle_graph(**kwargs: float) -> fraud_eagle.ReviewGraph:
        """Create a review graph defined in fraud eagle package."""
        if "epsilon" not in kwargs:
            LOGGER.warning("Parameter 'epsilon' is not specified. Set to 0.1.")
            kwargs["epsilon"] = 0.1
        return fraud_eagle.ReviewGraph(kwargs["epsilon"])

    ALGORITHMS["feagle"] = create_feagle_graph
except ImportError:
    LOGGER.info("rgmining-fraud-eagle is not installed.")

# Load and register FRAUDAR.
try:
    import fraudar

    def create_fraudar_graph(**kwargs: float) -> fraudar.ReviewGraph:
        """Create a review graph defined in Fraudar package."""
        if "nblock" not in kwargs:
            LOGGER.warning("Parameter 'nblock' is not specified. Set to 1.")
            kwargs["nblock"] = 1.0
        return fraudar.ReviewGraph(int(kwargs["nblock"]))

    ALGORITHMS["fraudar"] = create_fraudar_graph
except ImportError:
    LOGGER.info("rgmining-fraudar is not installed.")


def run(
    method: str, loop: int, threshold: float, output: TextIO, param: tuple[str]
) -> None:
    """Run a given algorithm with the Trip Advisor dataset.

    Runs a given algorithm and outputs anomalous scores and summaries after
    each iteration finishes. The function will end if a given number of loops
    ends or the update of one iteration becomes smaller than a given threshold.

    Some algorithm requires a set of parameters. For example, feagle requires
    parameter `epsilon`. Argument `param` specifies those parameters, and
    if you want to set 0.1 to the `epsilon`, pass `epsilon=0.1` via the
    argument.

    Args:
      method: name of algorithm.
      loop: the number of iteration (default: 20).
      threshold: threshold to judge an update is negligible (default: 10^-3).
      output: writable object where the output will be written.
      param: list of key and value pair which are connected with "=".
    """
    kwargs = {
        key: float(value) for key, value in [v.split("=") for v in param]
    }

    graph = ALGORITHMS[method](**kwargs)
    load(graph)

    print_state(graph, 0, output)

    # Updates
    LOGGER.info("Start iterations.")
    for i in range(loop if not method.startswith("one") else 1):
        diff = graph.update()
        if diff is not None and diff < threshold:
            break

        # Current summary
        LOGGER.info("Iteration %d ends. (diff=%s)", i + 1, diff)
        print_state(graph, i + 1, output)

    # Print final state.
    print_state(graph, "final", output)


@click.command()
@click.option(
    "-m",
    "--method",
    type=click.Choice(list(ALGORITHMS.keys()), case_sensitive=False),
    required=True,
    help="name of algorithm.",
)
@click.option("--loop", type=int, default=20, help="number of iteration.")
@click.option("--threshold", type=float, default=10 ^ -3, help="threshold.")
@click.option(
    "--output",
    default=sys.stdout,
    type=click.File("w"),
    help="file path to store results. [Default: stdout]",
)
@click.option(
    "--param",
    multiple=True,
    help="key and value pair of parameters corresponding to the chosen algorithm, connected with '='.",
)
@click.version_option(version("rgmining-tripadvisor-dataset"))
def main(
    method: str, loop: int, threshold: float, output: TextIO, param: tuple[str]
) -> None:
    """Evaluate a review graph mining algorithm with the Trip Advisor dataset."""
    logging.basicConfig(level=logging.INFO, stream=sys.stderr)
    run(method, loop, threshold, output, param)


__all__: Final = ["main"]



#
# loader.py
#
# Copyright (c) 2017-2025 Junpei Kawamoto
#
# This file is part of rgmining-tripadvisor-dataset.
#
# rgmining-tripadvisor-dataset is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# rgmining-tripadvisor-dataset is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Foobar.  If not, see <http://www.gnu.org/licenses/>.
#
"""This module provides a function to load the Trip Advisor dataset."""

import json
import logging
import tarfile
from collections.abc import Iterator
from contextlib import closing
from datetime import datetime
from typing import Any, BinaryIO, cast, Protocol, TypeVar

import requests
from platformdirs import user_cache_path
from tqdm import tqdm

LOGGER = logging.getLogger(__name__)

DATASET_URL = "https://www.cs.virginia.edu/~hw5x/Data/LARA/TripAdvisor/TripAdvisorJson.tar.bz2"
FILENAME = "TripAdvisorJson.tar.bz2"

_DATE_FORMAT = "%B %d, %Y"
"""Data format in the dataset.
"""

RT = TypeVar("RT")
PT = TypeVar("PT")


class Graph(Protocol[RT, PT]):
    """A protocol class representing a graph object.

    This protocol defines the methods required for a class to be used
    as the graph object in the `load` function. The graph object facilitates
    creating reviewers, products, and adding reviews.
    """

    def new_reviewer(
        self, name: str, anomalous_score: float | None = None
    ) -> RT:
        """Creates a new reviewer node in the graph.

        Args:
            name: The name of the reviewer.
            anomalous_score: Optional. The initial anomalous score for the reviewer.

        Returns:
            The created reviewer node.
        """

    def new_product(self, name: str) -> PT:
        """Creates a new product node in the graph.

        Args:
            name: The name of the product.

        Returns:
            The created product node.
        """

    def add_review(
        self, reviewer: RT, product: PT, score: float, time: int | None = None
    ) -> Any:
        """Adds a review connecting a reviewer to a product with a score and time.

        Args:
            reviewer: The reviewer node.
            product: The product node.
            score: The score of the review.
            time: Optional. The time of the review.
        """


def reviews() -> Iterator[dict[str, Any]]:
    """Load the Trip Advisor dataset."""

    data_path = user_cache_path(
        "rgmining-tripadvisor-dataset", ensure_exists=True
    ).joinpath(FILENAME)
    if not data_path.exists():
        LOGGER.info(
            "Not found review data locally, downloading them from %s...",
            DATASET_URL,
        )

        res = requests.get(DATASET_URL, stream=True)
        res.raise_for_status()
        with open(data_path, "wb") as f:
            for chunk in res.iter_content(chunk_size=32 * 1024):
                f.write(chunk)

        LOGGER.info("Downloaded review data are stored at %s", data_path)

    with tarfile.open(data_path) as tar:
        LOGGER.info("Extracting review data from %s...", data_path)
        for info in tqdm(tar.getmembers()):
            if not info.isfile():
                continue

            with closing(cast(BinaryIO, tar.extractfile(info))) as fp:
                yield json.load(fp)
                return


def load(graph: Graph) -> Graph:
    """Load the Trip Advisor dataset to a given graph object.

    Args:
      graph: an instance of review graph.

    Returns:
      The graph instance *graph*.
    """
    R = {}  # Reviewers dict.
    for obj in reviews():
        target = obj["HotelInfo"]["HotelID"]
        product = graph.new_product(name=target)

        for r in obj["Reviews"]:
            name = r["ReviewID"]
            score = float(r["Ratings"]["Overall"]) / 5.0

            try:
                date = int(
                    datetime.strptime(r["Date"], _DATE_FORMAT).strftime(
                        "%Y%m%d"
                    )
                )
            except ValueError:
                date = None

            if name not in R:
                R[name] = graph.new_reviewer(name=name)
            graph.add_review(R[name], product, score, date)

    return graph
